# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{wokkel.muc}
"""

from datetime import datetime
from dateutil.tz import tzutc

from zope.interface import verify

from twisted.trial import unittest
from twisted.internet import defer, task
from twisted.words.xish import domish, xpath
from twisted.words.protocols.jabber.jid import JID
from twisted.words.protocols.jabber.error import StanzaError
from twisted.words.protocols.jabber.xmlstream import TimeoutError, toResponse

from wokkel import data_form, delay, iwokkel, muc
from wokkel.generic import parseXml
from wokkel.test.helpers import TestableStreamManager


NS_MUC_ADMIN = 'http://jabber.org/protocol/muc#admin'

def calledAsync(fn):
    """
    Function wrapper that fires a deferred upon calling the given function.
    """
    d = defer.Deferred()

    def func(*args, **kwargs):
        try:
            result = fn(*args, **kwargs)
        except:
            d.errback()
        else:
            d.callback(result)

    return d, func



class StatusCodeTest(unittest.TestCase):
    """
    Tests for L{muc.STATUS_CODE}.
    """

    def test_lookupByValue(self):
        """
        The registered MUC status codes map to STATUS_CODE value constants.

        Note: the identifiers used in the dictionary of status codes are
        borrowed from U{XEP-0306<http://xmpp.org/extensions/xep-0306.html>}
        that defines Extensible Status Conditions for Multi-User Chat. If this
        specification is implemented itself, the dictionary could move there.
        """
        codes = {
            100: 'realjid-public',
            101: 'affiliation-changed',
            102: 'unavailable-shown',
            103: 'unavailable-not-shown',
            104: 'configuration-changed',
            110: 'self-presence',
            170: 'logging-enabled',
            171: 'logging-disabled',
            172: 'non-anonymous',
            173: 'semi-anonymous',
            174: 'fully-anonymous',
            201: 'room-created',
            210: 'nick-assigned',
            301: 'banned',
            303: 'new-nick',
            307: 'kicked',
            321: 'removed-affiliation',
            322: 'removed-membership',
            332: 'removed-shutdown',
        }

        for code, condition in codes.iteritems():
            constantName = condition.replace('-', '_').upper()
            self.assertEqual(getattr(muc.STATUS_CODE, constantName),
                             muc.STATUS_CODE.lookupByValue(code))



class StatusesTest(unittest.TestCase):
    """
    Tests for L{muc.Statuses}.
    """

    def setUp(self):
        self.mucStatuses = muc.Statuses()
        self.mucStatuses.add(muc.STATUS_CODE.SELF_PRESENCE)
        self.mucStatuses.add(muc.STATUS_CODE.ROOM_CREATED)


    def test_interface(self):
        """
        Instances of L{Statuses} provide L{iwokkel.IMUCStatuses}.
        """
        verify.verifyObject(iwokkel.IMUCStatuses, self.mucStatuses)


    def test_contains(self):
        """
        The status contained are 'in' the container.
        """
        self.assertIn(muc.STATUS_CODE.SELF_PRESENCE, self.mucStatuses)
        self.assertIn(muc.STATUS_CODE.ROOM_CREATED, self.mucStatuses)
        self.assertNotIn(muc.STATUS_CODE.NON_ANONYMOUS, self.mucStatuses)


    def test_iter(self):
        """
        All statuses can be iterated over.
        """
        statuses = set()
        for status in self.mucStatuses:
            statuses.add(status)

        self.assertEqual(set([muc.STATUS_CODE.SELF_PRESENCE,
                              muc.STATUS_CODE.ROOM_CREATED]), statuses)


    def test_len(self):
        """
        The number of items in this container is returned by C{__len__}.
        """
        self.assertEqual(2, len(self.mucStatuses))



class GroupChatTest(unittest.TestCase):
    """
    Tests for L{muc.GroupChat}.
    """


    def test_toElementDelay(self):
        """
        If the delay attribute is set, toElement has it rendered.
        """
        message = muc.GroupChat()
        message.delay = delay.Delay(stamp=datetime(2002, 10, 13, 23, 58, 37,
                                                   tzinfo=tzutc()))

        element = message.toElement()

        query = "/message/delay[@xmlns='%s']" % (delay.NS_DELAY,)
        nodes = xpath.queryForNodes(query, element)
        self.assertNotIdentical(None, nodes, "Missing delay element")


    def test_toElementDelayLegacy(self):
        """
        If legacy delay is requested, the legacy format is rendered.
        """
        message = muc.GroupChat()
        message.delay = delay.Delay(stamp=datetime(2002, 10, 13, 23, 58, 37,
                                                   tzinfo=tzutc()))

        element = message.toElement(legacyDelay=True)

        query = "/message/x[@xmlns='%s']" % (delay.NS_JABBER_DELAY,)
        nodes = xpath.queryForNodes(query, element)
        self.assertNotIdentical(None, nodes, "Missing legacy delay element")



class HistoryOptionsTest(unittest.TestCase):
    """
    Tests for L{muc.HistoryOptionsTest}.
    """

    def test_toElement(self):
        """
        toElement renders the history element in the right namespace.
        """
        history = muc.HistoryOptions()

        element = history.toElement()

        self.assertEqual(muc.NS_MUC, element.uri)
        self.assertEqual('history', element.name)


    def test_toElementMaxStanzas(self):
        """
        If C{maxStanzas} is set, the element has the attribute C{'maxstanzas'}.
        """
        history = muc.HistoryOptions(maxStanzas=10)

        element = history.toElement()

        self.assertEqual(u'10', element.getAttribute('maxstanzas'))


    def test_toElementSince(self):
        """
        If C{since} is set, the attribute C{'since'} has a rendered timestamp.
        """
        history = muc.HistoryOptions(since=datetime(2002, 10, 13, 23, 58, 37,
                                                   tzinfo=tzutc()))

        element = history.toElement()

        self.assertEqual(u'2002-10-13T23:58:37Z',
                         element.getAttribute('since'))



class UserPresenceTest(unittest.TestCase):
    """
    Tests for L{muc.UserPresence}.
    """

    def test_fromElementNoUserElement(self):
        """
        Without user element, all associated attributes are None.
        """
        xml = """
            <presence from='coven@chat.shakespeare.lit/thirdwitch'
                      id='026B3509-2CCE-4D69-96D6-25F41FFDC408'
                      to='hag66@shakespeare.lit/pda'>
            </presence>
        """

        element = parseXml(xml)
        presence = muc.UserPresence.fromElement(element)

        self.assertIdentical(None, presence.affiliation)
        self.assertIdentical(None, presence.role)
        self.assertIdentical(None, presence.entity)
        self.assertIdentical(None, presence.nick)
        self.assertEqual(0, len(presence.mucStatuses))


    def test_fromElementUnknownChild(self):
        """
        Unknown child elements are ignored.
        """
        xml = """
            <presence from='coven@chat.shakespeare.lit/thirdwitch'
                      id='026B3509-2CCE-4D69-96D6-25F41FFDC408'
                      to='hag66@shakespeare.lit/pda'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <status xmlns='myns' code='110'/>
              </x>
            </presence>
        """

        element = parseXml(xml)
        presence = muc.UserPresence.fromElement(element)

        self.assertEqual(0, len(presence.mucStatuses))


    def test_fromElementStatusOne(self):
        """
        Status codes are extracted.
        """
        xml = """
            <presence from='coven@chat.shakespeare.lit/thirdwitch'
                      id='026B3509-2CCE-4D69-96D6-25F41FFDC408'
                      to='hag66@shakespeare.lit/pda'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
                <status code='110'/>
              </x>
            </presence>
        """

        element = parseXml(xml)
        presence = muc.UserPresence.fromElement(element)

        self.assertIn(muc.STATUS_CODE.SELF_PRESENCE, presence.mucStatuses)


    def test_fromElementStatusMultiple(self):
        """
        Multiple status codes are all extracted.
        """
        xml = """
            <presence from='coven@chat.shakespeare.lit/thirdwitch'
                      id='026B3509-2CCE-4D69-96D6-25F41FFDC408'
                      to='hag66@shakespeare.lit/pda'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
                <status code='100'/>
                <status code='110'/>
              </x>
            </presence>
        """

        element = parseXml(xml)
        presence = muc.UserPresence.fromElement(element)

        self.assertIn(muc.STATUS_CODE.SELF_PRESENCE, presence.mucStatuses)
        self.assertIn(muc.STATUS_CODE.REALJID_PUBLIC, presence.mucStatuses)


    def test_fromElementStatusEmpty(self):
        """
        Empty status elements are ignored.
        """
        xml = """
            <presence from='coven@chat.shakespeare.lit/thirdwitch'
                      id='026B3509-2CCE-4D69-96D6-25F41FFDC408'
                      to='hag66@shakespeare.lit/pda'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
                <status/>
              </x>
            </presence>
        """

        element = parseXml(xml)
        presence = muc.UserPresence.fromElement(element)

        self.assertEqual(0, len(presence.mucStatuses))


    def test_fromElementStatusBad(self):
        """
        Bad status codes are ignored.
        """
        xml = """
            <presence from='coven@chat.shakespeare.lit/thirdwitch'
                      id='026B3509-2CCE-4D69-96D6-25F41FFDC408'
                      to='hag66@shakespeare.lit/pda'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
                <status code="badvalue"/>
              </x>
            </presence>
        """

        element = parseXml(xml)
        presence = muc.UserPresence.fromElement(element)

        self.assertEqual(0, len(presence.mucStatuses))


    def test_fromElementStatusUnknown(self):
        """
        Unknown status codes are not recorded in C{mucStatuses}.
        """
        xml = """
            <presence from='coven@chat.shakespeare.lit/thirdwitch'
                      id='026B3509-2CCE-4D69-96D6-25F41FFDC408'
                      to='hag66@shakespeare.lit/pda'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
                <status code="999"/>
              </x>
            </presence>
        """

        element = parseXml(xml)
        presence = muc.UserPresence.fromElement(element)

        self.assertEqual(0, len(presence.mucStatuses))


    def test_fromElementItem(self):
        """
        Item attributes are parsed properly.
        """
        xml = """
            <presence from='coven@chat.shakespeare.lit/thirdwitch'
                      to='crone1@shakespeare.lit/desktop'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member'
                      jid='hag66@shakespeare.lit/pda'
                      role='participant'
                      nick='thirdwitch'/>
              </x>
            </presence>
        """

        element = parseXml(xml)
        presence = muc.UserPresence.fromElement(element)
        self.assertEqual(u'member', presence.affiliation)
        self.assertEqual(u'participant', presence.role)
        self.assertEqual(JID('hag66@shakespeare.lit/pda'), presence.entity)
        self.assertEqual(u'thirdwitch', presence.nick)



class MUCClientProtocolTest(unittest.TestCase):
    """
    Tests for L{muc.MUCClientProtocol}.
    """

    def setUp(self):
        self.clock = task.Clock()
        self.sessionManager = TestableStreamManager(reactor=self.clock)
        self.stub = self.sessionManager.stub
        self.protocol = muc.MUCClientProtocol(reactor=self.clock)
        self.protocol.setHandlerParent(self.sessionManager)

        self.roomIdentifier = 'test'
        self.service  = 'conference.example.org'
        self.nick = 'Nick'

        self.occupantJID = JID(tuple=(self.roomIdentifier,
                                      self.service,
                                      self.nick))
        self.roomJID = self.occupantJID.userhostJID()
        self.userJID = JID('test@example.org/Testing')


    def test_initNoReactor(self):
        """
        If no reactor is passed, the default reactor is used.
        """
        protocol = muc.MUCClientProtocol()
        from twisted.internet import reactor
        self.assertEqual(reactor, protocol._reactor)


    def test_groupChatReceived(self):
        """
        Messages of type groupchat are parsed and passed to L{groupChatReceived}.
        """
        xml = u"""
            <message to='test@test.com' from='%s' type='groupchat'>
              <body>test</body>
            </message>
        """ % (self.occupantJID)

        def groupChatReceived(message):
            self.assertEquals('test', message.body, "Wrong group chat message")
            self.assertEquals(self.roomIdentifier, message.sender.user,
                              'Wrong room identifier')

        d, self.protocol.groupChatReceived = calledAsync(groupChatReceived)
        self.stub.send(parseXml(xml))
        return d


    def test_groupChatReceivedNotOverridden(self):
        """
        If L{groupChatReceived} has not been overridden, no errors should occur.
        """
        xml = u"""
            <message to='test@test.com' from='%s' type='groupchat'>
              <body>test</body>
            </message>
        """ % (self.occupantJID)

        self.stub.send(parseXml(xml))


    def test_join(self):
        """
        Joining a room waits for confirmation, deferred fires user presence.
        """

        def cb(presence):
            self.assertEquals(self.occupantJID, presence.sender)

        # Join the room
        d = self.protocol.join(self.roomJID, self.nick)
        d.addCallback(cb)

        element = self.stub.output[-1]

        self.assertEquals('presence', element.name, "Need to be presence")
        self.assertNotIdentical(None, element.x, 'No muc x element')

        # send back user presence, they joined
        xml = """
            <presence from='%s@%s/%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
              </x>
            </presence>
        """ % (self.roomIdentifier, self.service, self.nick)
        self.stub.send(parseXml(xml))
        return d


    def test_joinHistory(self):
        """
        Passing a history parameter sends a 'maxStanzas' history limit.
        """

        historyOptions = muc.HistoryOptions(maxStanzas=10)
        d = self.protocol.join(self.roomJID, self.nick,
                               historyOptions)

        element = self.stub.output[-1]
        query = "/*/x[@xmlns='%s']/history[@xmlns='%s']" % (muc.NS_MUC,
                                                            muc.NS_MUC)
        result = xpath.queryForNodes(query, element)
        history = result[0]
        self.assertEquals('10', history.getAttribute('maxstanzas'))

        # send back user presence, they joined
        xml = """
            <presence from='%s@%s/%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
              </x>
            </presence>
        """ % (self.roomIdentifier, self.service, self.nick)
        self.stub.send(parseXml(xml))
        return d


    def test_joinForbidden(self):
        """
        A forbidden error in response to a join errbacks with L{StanzaError}.
        """

        def cb(error):
            self.assertEquals('forbidden', error.condition,
                              'Wrong muc condition')

        d = self.protocol.join(self.roomJID, self.nick)
        self.assertFailure(d, StanzaError)
        d.addCallback(cb)

        # send back error, forbidden
        xml = u"""
            <presence from='%s' type='error'>
              <error type='auth'>
                <forbidden xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
              </error>
            </presence>
        """ % (self.occupantJID)
        self.stub.send(parseXml(xml))
        return d


    def test_joinForbiddenFromRoomJID(self):
        """
        An error response to a join sent from the room JID should errback.

        Some service implementations send error stanzas from the room JID
        instead of the JID the join presence was sent to.
        """

        d = self.protocol.join(self.roomJID, self.nick)
        self.assertFailure(d, StanzaError)

        # send back error, forbidden
        xml = u"""
            <presence from='%s' type='error'>
              <error type='auth'>
                <forbidden xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
              </error>
            </presence>
        """ % (self.roomJID)
        self.stub.send(parseXml(xml))
        return d


    def test_joinBadJID(self):
        """
        Client joining a room and getting a jid-malformed error.
        """

        def cb(error):
            self.assertEquals('jid-malformed', error.condition,
                              'Wrong muc condition')

        d = self.protocol.join(self.roomJID, self.nick)
        self.assertFailure(d, StanzaError)
        d.addCallback(cb)

        # send back error, bad JID
        xml = u"""
            <presence from='%s' type='error'>
              <error type='modify'>
                <jid-malformed xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
              </error>
            </presence>
        """ % (self.occupantJID)
        self.stub.send(parseXml(xml))
        return d


    def test_joinTimeout(self):
        """
        After not receiving a response to a join, errback with L{TimeoutError}.
        """

        d = self.protocol.join(self.roomJID, self.nick)
        self.assertFailure(d, TimeoutError)
        self.clock.advance(muc.DEFER_TIMEOUT)
        return d


    def test_joinPassword(self):
        """
        Sending a password via presence to a password protected room.
        """

        self.protocol.join(self.roomJID, self.nick, password='secret')

        element = self.stub.output[-1]

        self.assertTrue(xpath.matches(
                u"/presence[@to='%s']/x/password"
                    "[text()='secret']" % (self.occupantJID,),
                element),
            'Wrong presence stanza')


    def test_nick(self):
        """
        Send a nick change to the server.
        """
        newNick = 'newNick'

        def cb(presence):
            self.assertEquals(JID(tuple=(self.roomIdentifier,
                                         self.service,
                                         newNick)),
                              presence.sender)

        d = self.protocol.nick(self.roomJID, newNick)
        d.addCallback(cb)

        element = self.stub.output[-1]
        self.assertEquals('presence', element.name, "Need to be presence")
        self.assertNotIdentical(None, element.x, 'No muc x element')

        # send back user presence, nick changed
        xml = u"""
            <presence from='%s/%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
              </x>
            </presence>
        """ % (self.roomJID, newNick)
        self.stub.send(parseXml(xml))
        return d


    def test_nickConflict(self):
        """
        If the server finds the new nick in conflict, the errback is called.
        """
        newNick = 'newNick'

        d = self.protocol.nick(self.roomJID, newNick)
        self.assertFailure(d, StanzaError)

        element = self.stub.output[-1]
        self.assertEquals('presence', element.name, "Need to be presence")
        self.assertNotIdentical(None, element.x, 'No muc x element')

        # send back error presence, nick conflicted
        xml = u"""
            <presence from='%s/%s' type='error'>
                <x xmlns='http://jabber.org/protocol/muc'/>
                <error type='cancel'>
                  <conflict xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
                </error>
            </presence>
        """ % (self.roomJID, newNick)
        self.stub.send(parseXml(xml))
        return d


    def test_status(self):
        """
        Change status
        """
        def joined(_):
            d = self.protocol.status(self.roomJID, 'xa', 'testing MUC')
            d.addCallback(statusChanged)
            return d

        def statusChanged(presence):
            self.assertEqual(self.occupantJID, presence.sender)

        # Join the room
        d = self.protocol.join(self.roomJID, self.nick)
        d.addCallback(joined)

        # Receive presence back from the room: joined.
        xml = u"""
            <presence to='%s' from='%s'/>
        """ % (self.userJID, self.occupantJID)
        self.stub.send(parseXml(xml))

        # The presence for the status change should have been sent now.
        element = self.stub.output[-1]

        self.assertEquals('presence', element.name, "Need to be presence")
        self.assertTrue(getattr(element, 'x', None), 'No muc x element')

        # send back user presence, status changed
        xml = u"""
            <presence from='%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
              </x>
              <show>xa</show>
              <status>testing MUC</status>
            </presence>
        """ % self.occupantJID
        self.stub.send(parseXml(xml))

        return d


    def test_leave(self):
        """
        Client leaves a room
        """
        def joined(_):
            return self.protocol.leave(self.roomJID)

        # Join the room
        d = self.protocol.join(self.roomJID, self.nick)
        d.addCallback(joined)

        # Receive presence back from the room: joined.
        xml = u"""
            <presence to='%s' from='%s'/>
        """ % (self.userJID, self.occupantJID)
        self.stub.send(parseXml(xml))

        # The presence for leaving the room should have been sent now.
        element = self.stub.output[-1]

        self.assertEquals('unavailable', element['type'],
                          'Unavailable is not being sent')

        # Receive presence back from the room: left.
        xml = u"""
            <presence to='%s' from='%s' type='unavailable'/>
        """ % (self.userJID, self.occupantJID)
        self.stub.send(parseXml(xml))

        return d


    def test_groupChat(self):
        """
        Send private messages to muc entities.
        """
        self.protocol.groupChat(self.roomJID, u'This is a test')

        message = self.stub.output[-1]

        self.assertEquals('message', message.name)
        self.assertEquals(self.roomJID.full(), message.getAttribute('to'))
        self.assertEquals('groupchat', message.getAttribute('type'))
        self.assertEquals(u'This is a test', unicode(message.body))


    def test_chat(self):
        """
        Send private messages to muc entities.
        """
        otherOccupantJID = JID(self.occupantJID.userhost()+'/OtherNick')

        self.protocol.chat(otherOccupantJID, u'This is a test')

        message = self.stub.output[-1]

        self.assertEquals('message', message.name)
        self.assertEquals(otherOccupantJID.full(), message.getAttribute('to'))
        self.assertEquals('chat', message.getAttribute('type'))
        self.assertEquals(u'This is a test', unicode(message.body))


    def test_subject(self):
        """
        Change subject of the room.
        """
        self.protocol.subject(self.roomJID, u'This is a test')

        message = self.stub.output[-1]

        self.assertEquals('message', message.name)
        self.assertEquals(self.roomJID.full(), message.getAttribute('to'))
        self.assertEquals('groupchat', message.getAttribute('type'))
        self.assertEquals(u'This is a test', unicode(message.subject))


    def test_invite(self):
        """
        Invite a user to a room
        """
        invitee = JID('other@example.org')

        self.protocol.invite(self.roomJID, invitee, u'This is a test')

        message = self.stub.output[-1]

        self.assertEquals('message', message.name)
        self.assertEquals(self.roomJID.full(), message.getAttribute('to'))
        self.assertEquals(muc.NS_MUC_USER, message.x.uri)
        self.assertEquals(muc.NS_MUC_USER, message.x.invite.uri)
        self.assertEquals(invitee.full(), message.x.invite.getAttribute('to'))
        self.assertEquals(muc.NS_MUC_USER, message.x.invite.reason.uri)
        self.assertEquals(u'This is a test', unicode(message.x.invite.reason))


    def test_getRegisterForm(self):
        """
        The response of a register form request should extract the form.
        """

        def cb(form):
            self.assertEquals('form', form.formType)

        d = self.protocol.getRegisterForm(self.roomJID)
        d.addCallback(cb)

        iq = self.stub.output[-1]

        query = "/iq/query[@xmlns='%s']" % (muc.NS_REGISTER)
        nodes = xpath.queryForNodes(query, iq)
        self.assertNotIdentical(None, nodes, 'Missing query element')

        self.assertRaises(StopIteration, nodes[0].elements().next)

        xml = u"""
            <iq from='%s' id='%s' to='%s' type='result'>
              <query xmlns='jabber:iq:register'>
                <x xmlns='jabber:x:data' type='form'>
                  <field type='hidden'
                         var='FORM_TYPE'>
                    <value>http://jabber.org/protocol/muc#register</value>
                  </field>
                  <field label='Desired Nickname'
                         type='text-single'
                         var='muc#register_roomnick'>
                    <required/>
                  </field>
                </x>
              </query>
            </iq>
        """ % (self.roomJID, iq['id'], self.userJID)
        self.stub.send(parseXml(xml))

        return d


    def test_register(self):
        """
        Client registering with a room.

        http://xmpp.org/extensions/xep-0045.html#register
        """

        def cb(iq):
            # check for a result
            self.assertEquals('result', iq['type'], 'We did not get a result')

        d = self.protocol.register(self.roomJID,
                                   {'muc#register_roomnick': 'thirdwitch'})
        d.addCallback(cb)

        iq = self.stub.output[-1]

        query = "/iq/query[@xmlns='%s']" % muc.NS_REGISTER
        nodes = xpath.queryForNodes(query, iq)
        self.assertNotIdentical(None, nodes, 'Invalid registration request')

        form = data_form.findForm(nodes[0], muc.NS_MUC_REGISTER)
        self.assertNotIdentical(None, form, 'Missing registration form')
        self.assertEquals('submit', form.formType)
        self.assertIn('muc#register_roomnick', form.fields)


        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d


    def test_registerCancel(self):
        """
        Cancelling a registration request sends a cancel form.
        """

        d = self.protocol.register(self.roomJID, None)

        iq = self.stub.output[-1]

        query = "/iq/query[@xmlns='%s']" % muc.NS_REGISTER
        nodes = xpath.queryForNodes(query, iq)
        self.assertNotIdentical(None, nodes, 'Invalid registration request')

        form = data_form.findForm(nodes[0], muc.NS_MUC_REGISTER)
        self.assertNotIdentical(None, form, 'Missing registration form')
        self.assertEquals('cancel', form.formType)

        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d


    def test_voice(self):
        """
        Client requesting voice for a room.
        """
        self.protocol.voice(self.occupantJID)

        m = self.stub.output[-1]

        query = ("/message/x[@type='submit']/field/value"
                    "[text()='%s']") % muc.NS_MUC_REQUEST
        self.assertTrue(xpath.matches(query, m), 'Invalid voice message stanza')


    def test_history(self):
        """
        Converting a one to one chat to a multi-user chat.
        """
        archive = []
        thread = "e0ffe42b28561960c6b12b944a092794b9683a38"
        # create messages
        element = domish.Element((None, 'message'))
        element['to'] = 'testing@example.com'
        element['type'] = 'chat'
        element.addElement('body', None, 'test')
        element.addElement('thread', None, thread)

        archive.append({'stanza': element,
                        'timestamp': datetime(2002, 10, 13, 23, 58, 37,
                                              tzinfo=tzutc())})

        element = domish.Element((None, 'message'))
        element['to'] = 'testing2@example.com'
        element['type'] = 'chat'
        element.addElement('body', None, 'yo')
        element.addElement('thread', None, thread)

        archive.append({'stanza': element,
                        'timestamp': datetime(2002, 10, 13, 23, 58, 43,
                                              tzinfo=tzutc())})

        self.protocol.history(self.occupantJID, archive)


        while len(self.stub.output)>0:
            element = self.stub.output.pop()
            # check for delay element
            self.assertEquals('message', element.name, 'Wrong stanza')
            self.assertTrue(xpath.matches("/message/delay", element),
                            'Invalid history stanza')


    def test_getConfiguration(self):
        """
        The response of a configure form request should extract the form.
        """

        def cb(form):
            self.assertEquals('form', form.formType)

        d = self.protocol.getConfiguration(self.roomJID)
        d.addCallback(cb)

        iq = self.stub.output[-1]

        query = "/iq/query[@xmlns='%s']" % (muc.NS_MUC_OWNER)
        nodes = xpath.queryForNodes(query, iq)
        self.assertNotIdentical(None, nodes, 'Missing query element')

        self.assertRaises(StopIteration, nodes[0].elements().next)

        xml = u"""
            <iq from='%s' id='%s' to='%s' type='result'>
              <query xmlns='http://jabber.org/protocol/muc#owner'>
                <x xmlns='jabber:x:data' type='form'>
                  <field type='hidden'
                         var='FORM_TYPE'>
                    <value>http://jabber.org/protocol/muc#roomconfig</value>
                  </field>
                  <field label='Natural-Language Room Name'
                         type='text-single'
                         var='muc#roomconfig_roomname'/>
                </x>
              </query>
            </iq>
        """ % (self.roomJID, iq['id'], self.userJID)
        self.stub.send(parseXml(xml))

        return d


    def test_getConfigurationNoOptions(self):
        """
        The response of a configure form request should extract the form.
        """

        def cb(form):
            self.assertIdentical(None, form)

        d = self.protocol.getConfiguration(self.roomJID)
        d.addCallback(cb)

        iq = self.stub.output[-1]

        xml = u"""
            <iq from='%s' id='%s' to='%s' type='result'>
              <query xmlns='http://jabber.org/protocol/muc#owner'/>
            </iq>
        """ % (self.roomJID, iq['id'], self.userJID)
        self.stub.send(parseXml(xml))

        return d


    def test_configure(self):
        """
        Default configure and changing the room name.
        """

        def cb(iq):
            self.assertEquals('result', iq['type'], 'Not a result')

        values = {'muc#roomconfig_roomname': self.roomIdentifier}

        d = self.protocol.configure(self.roomJID, values)
        d.addCallback(cb)

        iq = self.stub.output[-1]

        self.assertEquals('set', iq.getAttribute('type'))
        self.assertEquals(self.roomJID.full(), iq.getAttribute('to'))

        query = "/iq/query[@xmlns='%s']" % (muc.NS_MUC_OWNER)
        nodes = xpath.queryForNodes(query, iq)
        self.assertNotIdentical(None, nodes, 'Bad configure request')

        form = data_form.findForm(nodes[0], muc.NS_MUC_CONFIG)
        self.assertNotIdentical(None, form, 'Missing configuration form')
        self.assertEquals('submit', form.formType)

        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d


    def test_configureEmpty(self):
        """
        Accept default configuration by sending an empty form.
        """

        values = {}

        d = self.protocol.configure(self.roomJID, values)

        iq = self.stub.output[-1]

        query = "/iq/query[@xmlns='%s']" % (muc.NS_MUC_OWNER)
        nodes = xpath.queryForNodes(query, iq)

        form = data_form.findForm(nodes[0], muc.NS_MUC_CONFIG)
        self.assertNotIdentical(None, form, 'Missing configuration form')
        self.assertEquals('submit', form.formType)

        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d


    def test_configureCancel(self):
        """
        Cancelling room configuration should send a cancel form.
        """

        d = self.protocol.configure(self.roomJID, None)

        iq = self.stub.output[-1]

        query = "/iq/query[@xmlns='%s']" % (muc.NS_MUC_OWNER)
        nodes = xpath.queryForNodes(query, iq)

        form = data_form.findForm(nodes[0], muc.NS_MUC_CONFIG)
        self.assertNotIdentical(None, form, 'Missing configuration form')
        self.assertEquals('cancel', form.formType)

        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d


    def test_getMemberList(self):
        """
        Retrieving the member list returns a list of L{muc.AdminItem}s

        The request asks for the affiliation C{'member'}.
        """
        def cb(items):
            self.assertEquals(1, len(items))
            item = items[0]
            self.assertEquals(JID(u'hag66@shakespeare.lit'), item.entity)
            self.assertEquals(u'thirdwitch', item.nick)
            self.assertEquals(u'member', item.affiliation)

        d = self.protocol.getMemberList(self.roomJID)
        d.addCallback(cb)

        iq = self.stub.output[-1]
        self.assertEquals('get', iq.getAttribute('type'))
        query = "/iq/query[@xmlns='%s']/item[@xmlns='%s']" % (muc.NS_MUC_ADMIN,
                                                              muc.NS_MUC_ADMIN)
        items = xpath.queryForNodes(query, iq)
        self.assertNotIdentical(None, items)
        self.assertEquals(1, len(items))
        self.assertEquals('member', items[0].getAttribute('affiliation'))

        response = toResponse(iq, 'result')
        query = response.addElement((NS_MUC_ADMIN, 'query'))
        item = query.addElement('item')
        item['affiliation'] ='member'
        item['jid'] = 'hag66@shakespeare.lit'
        item['nick'] = 'thirdwitch'
        item['role'] = 'participant'
        self.stub.send(response)

        return d


    def test_getAdminList(self):
        """
        Retrieving the admin list returns a list of L{muc.AdminItem}s

        The request asks for the affiliation C{'admin'}.
        """
        d = self.protocol.getAdminList(self.roomJID)

        iq = self.stub.output[-1]
        query = "/iq/query[@xmlns='%s']/item[@xmlns='%s']" % (muc.NS_MUC_ADMIN,
                                                              muc.NS_MUC_ADMIN)
        items = xpath.queryForNodes(query, iq)
        self.assertEquals('admin', items[0].getAttribute('affiliation'))

        response = toResponse(iq, 'result')
        query = response.addElement((NS_MUC_ADMIN, 'query'))
        self.stub.send(response)

        return d


    def test_getBanList(self):
        """
        Retrieving the ban list returns a list of L{muc.AdminItem}s

        The request asks for the affiliation C{'outcast'}.
        """
        def cb(items):
            self.assertEquals(1, len(items))
            item = items[0]
            self.assertEquals(JID(u'hag66@shakespeare.lit'), item.entity)
            self.assertEquals(u'outcast', item.affiliation)
            self.assertEquals(u'Trouble making', item.reason)

        d = self.protocol.getBanList(self.roomJID)
        d.addCallback(cb)

        iq = self.stub.output[-1]
        query = "/iq/query[@xmlns='%s']/item[@xmlns='%s']" % (muc.NS_MUC_ADMIN,
                                                              muc.NS_MUC_ADMIN)
        items = xpath.queryForNodes(query, iq)
        self.assertEquals('outcast', items[0].getAttribute('affiliation'))

        response = toResponse(iq, 'result')
        query = response.addElement((NS_MUC_ADMIN, 'query'))
        item = query.addElement('item')
        item['affiliation'] ='outcast'
        item['jid'] = 'hag66@shakespeare.lit'
        item.addElement('reason', content='Trouble making')
        self.stub.send(response)

        return d


    def test_getOwnerList(self):
        """
        Retrieving the owner list returns a list of L{muc.AdminItem}s

        The request asks for the affiliation C{'owner'}.
        """
        d = self.protocol.getOwnerList(self.roomJID)

        iq = self.stub.output[-1]
        query = "/iq/query[@xmlns='%s']/item[@xmlns='%s']" % (muc.NS_MUC_ADMIN,
                                                              muc.NS_MUC_ADMIN)
        items = xpath.queryForNodes(query, iq)
        self.assertEquals('owner', items[0].getAttribute('affiliation'))

        response = toResponse(iq, 'result')
        query = response.addElement((NS_MUC_ADMIN, 'query'))
        self.stub.send(response)

        return d


    def test_getModeratorList(self):
        """
        Retrieving the moderator returns a list of L{muc.AdminItem}s.

        The request asks for the role C{'moderator'}.
        """

        def cb(items):
            self.assertEquals(1, len(items))
            item = items[0]
            self.assertEquals(JID(u'hag66@shakespeare.lit'), item.entity)
            self.assertEquals(u'thirdwitch', item.nick)
            self.assertEquals(u'moderator', item.role)

        d = self.protocol.getModeratorList(self.roomJID)
        d.addCallback(cb)

        iq = self.stub.output[-1]
        self.assertEquals('get', iq.getAttribute('type'))
        query = "/iq/query[@xmlns='%s']/item[@xmlns='%s']" % (muc.NS_MUC_ADMIN,
                                                              muc.NS_MUC_ADMIN)
        items = xpath.queryForNodes(query, iq)
        self.assertNotIdentical(None, items)
        self.assertEquals(1, len(items))
        self.assertEquals('moderator', items[0].getAttribute('role'))

        response = toResponse(iq, 'result')
        query = response.addElement((NS_MUC_ADMIN, 'query'))
        item = query.addElement('item')
        item['affiliation'] ='member'
        item['jid'] = 'hag66@shakespeare.lit'
        item['nick'] = 'thirdwitch'
        item['role'] = 'moderator'
        self.stub.send(response)

        return d


    def test_modifyAffiliationList(self):

        entities = [JID('user1@test.example.org'),
                    JID('user2@test.example.org')]
        d = self.protocol.modifyAffiliationList(self.roomJID, entities,
                                                'admin')

        iq = self.stub.output[-1]
        query = "/iq/query[@xmlns='%s']/item[@xmlns='%s']" % (muc.NS_MUC_ADMIN,
                                                              muc.NS_MUC_ADMIN)
        items = xpath.queryForNodes(query, iq)
        self.assertNotIdentical(None, items)
        self.assertEquals(entities[0], JID(items[0].getAttribute('jid')))
        self.assertEquals('admin', items[0].getAttribute('affiliation'))
        self.assertEquals(entities[1], JID(items[1].getAttribute('jid')))
        self.assertEquals('admin', items[1].getAttribute('affiliation'))

        # Send a response to have the deferred fire.
        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d


    def test_grantVoice(self):
        """
        Granting voice sends request to set role to 'participant'.
        """
        nick = 'TroubleMaker'
        def cb(give_voice):
            self.assertTrue(give_voice, 'Did not give voice user')

        d = self.protocol.grantVoice(self.roomJID, nick,
                                     sender=self.userJID)
        d.addCallback(cb)

        iq = self.stub.output[-1]

        query = (u"/iq[@type='set' and @to='%s']/query/item"
                     "[@role='participant']") % self.roomJID
        self.assertTrue(xpath.matches(query, iq), 'Wrong voice stanza')

        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d


    def test_revokeVoice(self):
        """
        Revoking voice sends request to set role to 'visitor'.
        """
        nick = 'TroubleMaker'

        d = self.protocol.revokeVoice(self.roomJID, nick,
                                      reason="Trouble maker",
                                      sender=self.userJID)

        iq = self.stub.output[-1]

        query = (u"/iq[@type='set' and @to='%s']/query/item"
                     "[@role='visitor']") % self.roomJID
        self.assertTrue(xpath.matches(query, iq), 'Wrong voice stanza')

        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d


    def test_grantModerator(self):
        """
        Granting moderator privileges sends request to set role to 'moderator'.
        """
        nick = 'TroubleMaker'

        d = self.protocol.grantModerator(self.roomJID, nick,
                                         sender=self.userJID)

        iq = self.stub.output[-1]

        query = (u"/iq[@type='set' and @to='%s']/query/item"
                     "[@role='moderator']") % self.roomJID
        self.assertTrue(xpath.matches(query, iq), 'Wrong voice stanza')

        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d


    def test_ban(self):
        """
        Ban an entity in a room.
        """
        banned = JID('ban@jabber.org/TroubleMaker')

        def cb(banned):
            self.assertTrue(banned, 'Did not ban user')

        d = self.protocol.ban(self.roomJID, banned, reason='Spam',
                              sender=self.userJID)
        d.addCallback(cb)

        iq = self.stub.output[-1]

        self.assertTrue(xpath.matches(
                u"/iq[@type='set' and @to='%s']/query/item"
                    "[@affiliation='outcast']" % (self.roomJID,),
                iq),
            'Wrong ban stanza')

        response = toResponse(iq, 'result')
        self.stub.send(response)

        return d


    def test_kick(self):
        """
        Kick an entity from a room.
        """
        nick = 'TroubleMaker'

        def cb(kicked):
            self.assertTrue(kicked, 'Did not kick user')

        d = self.protocol.kick(self.roomJID, nick, reason='Spam',
                               sender=self.userJID)
        d.addCallback(cb)

        iq = self.stub.output[-1]

        self.assertTrue(xpath.matches(
                u"/iq[@type='set' and @to='%s']/query/item"
                    "[@role='none']" % (self.roomJID,),
                iq),
            'Wrong kick stanza')

        response = toResponse(iq, 'result')
        self.stub.send(response)

        return d


    def test_destroy(self):
        """
        Destroy a room.
        """
        d = self.protocol.destroy(self.occupantJID, reason='Time to leave',
                                  alternate=JID('other@%s' % self.service),
                                  password='secret')

        iq = self.stub.output[-1]

        query = ("/iq/query[@xmlns='%s']/destroy[@xmlns='%s']" %
                 (muc.NS_MUC_OWNER, muc.NS_MUC_OWNER))

        nodes = xpath.queryForNodes(query, iq)
        self.assertNotIdentical(None, nodes, 'Bad configure request')
        destroy = nodes[0]
        self.assertEquals('Time to leave', unicode(destroy.reason))

        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d



class MUCClientTest(unittest.TestCase):
    """
    Tests for C{muc.MUCClient}.
    """

    def setUp(self):
        self.clock = task.Clock()
        self.sessionManager = TestableStreamManager(reactor=self.clock)
        self.stub = self.sessionManager.stub
        self.protocol = muc.MUCClient(reactor=self.clock)
        self.protocol.setHandlerParent(self.sessionManager)

        self.roomIdentifier = 'test'
        self.service  = 'conference.example.org'
        self.nick = 'Nick'

        self.occupantJID = JID(tuple=(self.roomIdentifier,
                                      self.service,
                                      self.nick))
        self.roomJID = self.occupantJID.userhostJID()
        self.userJID = JID('test@example.org/Testing')


    def _createRoom(self):
        """
        A helper method to create a test room.
        """
        # create a room
        room = muc.Room(self.roomJID, self.nick)
        self.protocol._addRoom(room)
        return room


    def test_interface(self):
        """
        Do instances of L{muc.MUCClient} provide L{iwokkel.IMUCClient}?
        """
        verify.verifyObject(iwokkel.IMUCClient, self.protocol)


    def _testPresence(self, sender='', available=True):
        """
        Helper for presence tests.
        """
        def userUpdatedStatus(room, user, show, status):
            self.fail("Unexpected call to userUpdatedStatus")

        def userJoinedRoom(room, user):
            self.fail("Unexpected call to userJoinedRoom")

        if available:
            available = ""
        else:
            available = " type='unavailable'"

        if sender:
            sender = u" from='%s'" % sender

        xml = u"""
            <presence to='%s'%s%s>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
              </x>
            </presence>
        """ % (self.userJID, sender, available)

        self.protocol.userUpdatedStatus = userUpdatedStatus
        self.protocol.userJoinedRoom = userJoinedRoom
        self.stub.send(parseXml(xml))


    def test_availableReceivedEmptySender(self):
        """
        Availability presence from empty sender is ignored.
        """
        self._testPresence(sender='')


    def test_availableReceivedNotInRoom(self):
        """
        Availability presence from unknown entities is ignored.
        """
        otherOccupantJID = JID(self.occupantJID.userhost()+'/OtherNick')
        self._testPresence(sender=otherOccupantJID)


    def test_unavailableReceivedEmptySender(self):
        """
        Availability presence from empty sender is ignored.
        """
        self._testPresence(sender='', available=False)


    def test_unavailableReceivedNotInRoom(self):
        """
        Availability presence from unknown entities is ignored.
        """
        otherOccupantJID = JID(self.occupantJID.userhost()+'/OtherNick')
        self._testPresence(sender=otherOccupantJID, available=False)


    def test_unavailableReceivedNotInRoster(self):
        """
        Availability presence from unknown entities is ignored.
        """
        room = self._createRoom()
        user = muc.User(self.nick)
        room.addUser(user)
        otherOccupantJID = JID(self.occupantJID.userhost()+'/OtherNick')
        self._testPresence(sender=otherOccupantJID, available=False)


    def test_userJoinedRoom(self):
        """
        Joins by others to a room we're in are passed to userJoinedRoom
        """
        xml = """
            <presence to='%s' from='%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
              </x>
            </presence>
        """ % (self.userJID.full(), self.occupantJID.full())

        # create a room
        self._createRoom()

        def userJoinedRoom(room, user):
            self.assertEquals(self.roomJID, room.roomJID,
                              'Wrong room name')
            self.assertTrue(room.inRoster(user), 'User not in roster')

        d, self.protocol.userJoinedRoom = calledAsync(userJoinedRoom)
        self.stub.send(parseXml(xml))
        return d


    def test_receivedSubject(self):
        """
        Subject received from a room we're in are passed to receivedSubject.
        """
        xml = u"""
            <message to='%s' from='%s' type='groupchat'>
              <subject>test</subject>
            </message>
        """ % (self.userJID, self.occupantJID)

        self._createRoom()

        # add user to room
        user = muc.User(self.nick)
        room = self.protocol._getRoom(self.roomJID)
        room.addUser(user)

        def receivedSubject(room, user, subject):
            self.assertEquals('test', subject, "Wrong group chat message")
            self.assertEquals(self.roomJID, room.roomJID,
                              'Wrong room name')
            self.assertEquals(self.nick, user.nick)

        d, self.protocol.receivedSubject = calledAsync(receivedSubject)
        self.stub.send(parseXml(xml))
        return d


    def test_receivedSubjectNotOverridden(self):
        """
        Not overriding receivedSubject is ok.
        """
        xml = u"""
            <message to='%s' from='%s' type='groupchat'>
              <subject>test</subject>
            </message>
        """ % (self.userJID, self.occupantJID)

        self._createRoom()
        self.stub.send(parseXml(xml))


    def test_receivedGroupChat(self):
        """
        Messages received from a room we're in are passed to receivedGroupChat.
        """
        xml = u"""
            <message to='test@test.com' from='%s' type='groupchat'>
              <body>test</body>
            </message>
        """ % (self.occupantJID)

        self._createRoom()

        def receivedGroupChat(room, user, message):
            self.assertEquals('test', message.body, "Wrong group chat message")
            self.assertEquals(self.roomJID, room.roomJID,
                              'Wrong room name')

        d, self.protocol.receivedGroupChat = calledAsync(receivedGroupChat)
        self.stub.send(parseXml(xml))
        return d


    def test_receivedGroupChatRoom(self):
        """
        Messages received from the room itself have C{user} set to C{None}.
        """
        xml = u"""
            <message to='test@test.com' from='%s' type='groupchat'>
              <body>test</body>
            </message>
        """ % (self.roomJID)

        self._createRoom()

        def receivedGroupChat(room, user, message):
            self.assertIdentical(None, user)

        d, self.protocol.receivedGroupChat = calledAsync(receivedGroupChat)
        self.stub.send(parseXml(xml))
        return d


    def test_receivedGroupChatNotInRoom(self):
        """
        Messages received from a room we're not in are ignored.
        """
        xml = u"""
            <message to='test@test.com' from='%s' type='groupchat'>
              <body>test</body>
            </message>
        """ % (self.occupantJID)

        def receivedGroupChat(room, user, message):
            self.fail("Unexpected call to receivedGroupChat")

        self.protocol.receivedGroupChat = receivedGroupChat
        self.stub.send(parseXml(xml))


    def test_receivedGroupChatNotOverridden(self):
        """
        Not overriding receivedGroupChat is ok.
        """
        xml = u"""
            <message to='test@test.com' from='%s' type='groupchat'>
              <body>test</body>
            </message>
        """ % (self.occupantJID)

        self._createRoom()
        self.stub.send(parseXml(xml))


    def test_join(self):
        """
        Joining a room waits for confirmation, deferred fires room.
        """

        def cb(room):
            self.assertEqual(self.roomJID, room.roomJID)
            self.assertFalse(room.locked)

        d = self.protocol.join(self.roomJID, self.nick)
        d.addCallback(cb)

        # send back user presence, they joined
        xml = """
            <presence from='%s@%s/%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
              </x>
            </presence>
        """ % (self.roomIdentifier, self.service, self.nick)
        self.stub.send(parseXml(xml))
        return d


    def test_joinLocked(self):
        """
        A new room is locked by default.
        """

        def cb(room):
            self.assertTrue(room.locked, "Room is not marked as locked")

        d = self.protocol.join(self.roomJID, self.nick)
        d.addCallback(cb)

        # send back user presence, they joined
        xml = """
            <presence from='%s@%s/%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='owner' role='moderator'/>
                <status code="110"/>
                <status code="201"/>
              </x>
            </presence>
        """ % (self.roomIdentifier, self.service, self.nick)
        self.stub.send(parseXml(xml))
        return d


    def test_joinForbidden(self):
        """
        A forbidden error in response to a join errbacks with L{StanzaError}.
        """

        def cb(error):
            self.assertEquals('forbidden', error.condition,
                              'Wrong muc condition')
            self.assertIdentical(None, self.protocol._getRoom(self.roomJID))


        d = self.protocol.join(self.roomJID, self.nick)
        self.assertFailure(d, StanzaError)
        d.addCallback(cb)

        # send back error, forbidden
        xml = u"""
            <presence from='%s' type='error'>
              <error type='auth'>
                <forbidden xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
              </error>
            </presence>
        """ % (self.occupantJID)
        self.stub.send(parseXml(xml))
        return d


    def test_userLeftRoom(self):
        """
        Unavailable presence from a participant removes it from the room.
        """

        xml = u"""
            <presence to='%s' from='%s' type='unavailable'/>
        """ % (self.userJID, self.occupantJID)

        # create a room
        self._createRoom()

        # add user to room
        user = muc.User(self.nick)
        room = self.protocol._getRoom(self.roomJID)
        room.addUser(user)

        def userLeftRoom(room, user):
            self.assertEquals(self.roomJID, room.roomJID,
                              'Wrong room name')
            self.assertFalse(room.inRoster(user), 'User in roster')

        d, self.protocol.userLeftRoom = calledAsync(userLeftRoom)
        self.stub.send(parseXml(xml))
        return d


    def test_receivedHistory(self):
        """
        Receiving history on room join.
        """
        xml = u"""
            <message to='test@test.com' from='%s' type='groupchat'>
              <body>test</body>
              <delay xmlns='urn:xmpp:delay' stamp="2002-10-13T23:58:37Z"
                                            from="%s"/>
            </message>
        """ % (self.occupantJID, self.userJID)

        self._createRoom()


        def receivedHistory(room, user, message):
            self.assertEquals('test', message.body, "wrong message body")
            stamp = datetime(2002, 10, 13, 23, 58, 37, tzinfo=tzutc())
            self.assertEquals(stamp, message.delay.stamp,
                             'Does not have a history stamp')

        d, self.protocol.receivedHistory = calledAsync(receivedHistory)
        self.stub.send(parseXml(xml))
        return d


    def test_receivedHistoryNotOverridden(self):
        """
        Not overriding receivedHistory is ok.
        """
        xml = u"""
            <message to='test@test.com' from='%s' type='groupchat'>
              <body>test</body>
              <delay xmlns='urn:xmpp:delay' stamp="2002-10-13T23:58:37Z"
                                            from="%s"/>
            </message>
        """ % (self.occupantJID, self.userJID)

        self._createRoom()
        self.stub.send(parseXml(xml))


    def test_nickConflict(self):
        """
        If the server finds the new nick in conflict, the errback is called.
        """

        def cb(failure, room):
            user = room.getUser(otherNick)
            self.assertNotIdentical(None, user)
            self.assertEqual(otherJID, user.entity)

        def joined(room):
            d = self.protocol.nick(room.roomJID, otherNick)
            self.assertFailure(d, StanzaError)
            d.addCallback(cb, room)

        otherJID = JID('other@example.org/Home')
        otherNick = 'otherNick'

        d = self.protocol.join(self.roomJID, self.nick)
        d.addCallback(joined)

        # Send back other partipant's presence.
        xml = u"""
            <presence from='%s/%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant' jid='%s'/>
              </x>
            </presence>
        """ % (self.roomJID, otherNick, otherJID)
        self.stub.send(parseXml(xml))

        # send back user presence, they joined
        xml = u"""
            <presence from='%s/%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
              </x>
            </presence>
        """ % (self.roomJID, self.nick)
        self.stub.send(parseXml(xml))

        # send back error presence, nick conflicted
        xml = u"""
            <presence from='%s/%s' type='error'>
                <x xmlns='http://jabber.org/protocol/muc'/>
                <error type='cancel'>
                  <conflict xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
                </error>
            </presence>
        """ % (self.roomJID, otherNick)
        self.stub.send(parseXml(xml))
        return d


    def test_nick(self):
        """
        Send a nick change to the server.
        """
        newNick = 'newNick'

        room = self._createRoom()

        def joined(room):
            self.assertEqual(self.roomJID, room.roomJID)
            self.assertEqual(newNick, room.nick)
            user = room.getUser(newNick)
            self.assertNotIdentical(None, user)
            self.assertEqual(newNick, user.nick)

        d = self.protocol.nick(self.roomJID, newNick)
        d.addCallback(joined)

        # Nick should not have been changed, yet, as we haven't gotten
        # confirmation, yet.

        self.assertEquals(self.nick, room.nick)

        # send back user presence, nick changed
        xml = u"""
            <presence from='%s/%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
              </x>
            </presence>
        """ % (self.roomJID, newNick)

        self.stub.send(parseXml(xml))
        return d


    def test_leave(self):
        """
        Client leaves a room
        """
        def joined(_):
            return self.protocol.leave(self.roomJID)

        def left(_):
            self.assertIdentical(None, self.protocol._getRoom(self.roomJID))

        # Join the room
        d = self.protocol.join(self.roomJID, self.nick)
        d.addCallback(joined)
        d.addCallback(left)

        # Receive presence back from the room: joined.
        xml = u"""
            <presence to='%s' from='%s'/>
        """ % (self.userJID, self.occupantJID)
        self.stub.send(parseXml(xml))

        # Receive presence back from the room: left.
        xml = u"""
            <presence to='%s' from='%s' type='unavailable'/>
        """ % (self.userJID, self.occupantJID)
        self.stub.send(parseXml(xml))

        return d


    def test_status(self):
        """
        Change status
        """
        def joined(_):
            d = self.protocol.status(self.roomJID, 'xa', 'testing MUC')
            d.addCallback(statusChanged)
            return d

        def statusChanged(room):
            self.assertEqual(self.roomJID, room.roomJID)
            user = room.getUser(self.nick)
            self.assertNotIdentical(None, user, 'User not found')
            self.assertEqual('testing MUC', user.status, 'Wrong status')
            self.assertEqual('xa', user.show, 'Wrong show')

        # Join the room
        d = self.protocol.join(self.roomJID, self.nick)
        d.addCallback(joined)

        # Receive presence back from the room: joined.
        xml = u"""
            <presence to='%s' from='%s'/>
        """ % (self.userJID, self.occupantJID)
        self.stub.send(parseXml(xml))

        # send back user presence, status changed
        xml = u"""
            <presence from='%s'>
              <x xmlns='http://jabber.org/protocol/muc#user'>
                <item affiliation='member' role='participant'/>
              </x>
              <show>xa</show>
              <status>testing MUC</status>
            </presence>
        """ % self.occupantJID

        self.stub.send(parseXml(xml))
        return d


    def test_destroy(self):
        """
        Destroy a room.
        """
        def destroyed(_):
            self.assertIdentical(None, self.protocol._getRoom(self.roomJID))

        d = self.protocol.destroy(self.occupantJID, reason='Time to leave',
                                  alternate=JID('other@%s' % self.service),
                                  password='secret')
        d.addCallback(destroyed)

        iq = self.stub.output[-1]
        response = toResponse(iq, 'result')
        self.stub.send(response)
        return d



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{wokkel.disco}.
"""

from zope.interface import implements

from twisted.internet import defer
from twisted.trial import unittest
from twisted.words.protocols.jabber.error import StanzaError
from twisted.words.protocols.jabber.jid import JID
from twisted.words.protocols.jabber.xmlstream import toResponse
from twisted.words.xish import domish, utility

from wokkel import data_form, disco
from wokkel.generic import parseXml
from wokkel.subprotocols import XMPPHandler
from wokkel.test.helpers import TestableRequestHandlerMixin, XmlStreamStub

NS_DISCO_INFO = 'http://jabber.org/protocol/disco#info'
NS_DISCO_ITEMS = 'http://jabber.org/protocol/disco#items'

class DiscoFeatureTest(unittest.TestCase):
    """
    Tests for L{disco.DiscoFeature}.
    """

    def test_init(self):
        """
        Test initialization with a with feature namespace URI.
        """
        feature = disco.DiscoFeature(u'testns')
        self.assertEqual(u'testns', feature)


    def test_toElement(self):
        """
        Test proper rendering to a DOM representation.

        The returned element should be properly named and have a C{var}
        attribute that holds the feature namespace URI.
        """
        feature = disco.DiscoFeature(u'testns')
        element = feature.toElement()
        self.assertEqual(NS_DISCO_INFO, element.uri)
        self.assertEqual(u'feature', element.name)
        self.assertTrue(element.hasAttribute(u'var'))
        self.assertEqual(u'testns', element[u'var'])


    def test_fromElement(self):
        """
        Test creating L{disco.DiscoFeature} from L{domish.Element}.
        """
        element = domish.Element((NS_DISCO_INFO, u'feature'))
        element['var'] = u'testns'
        feature = disco.DiscoFeature.fromElement(element)
        self.assertEqual(u'testns', feature)



class DiscoIdentityTest(unittest.TestCase):
    """
    Tests for L{disco.DiscoIdentity}.
    """

    def test_init(self):
        """
        Test initialization with a category, type and name.
        """
        identity = disco.DiscoIdentity(u'conference', u'text', u'The chatroom')
        self.assertEqual(u'conference', identity.category)
        self.assertEqual(u'text', identity.type)
        self.assertEqual(u'The chatroom', identity.name)


    def test_toElement(self):
        """
        Test proper rendering to a DOM representation.

        The returned element should be properly named and have C{conference},
        C{type}, and C{name} attributes.
        """
        identity = disco.DiscoIdentity(u'conference', u'text', u'The chatroom')
        element = identity.toElement()
        self.assertEqual(NS_DISCO_INFO, element.uri)
        self.assertEqual(u'identity', element.name)
        self.assertEqual(u'conference', element.getAttribute(u'category'))
        self.assertEqual(u'text', element.getAttribute(u'type'))
        self.assertEqual(u'The chatroom', element.getAttribute(u'name'))


    def test_toElementWithoutName(self):
        """
        Test proper rendering to a DOM representation without a name.

        The returned element should be properly named and have C{conference},
        C{type} attributes, no C{name} attribute.
        """
        identity = disco.DiscoIdentity(u'conference', u'text')
        element = identity.toElement()
        self.assertEqual(NS_DISCO_INFO, element.uri)
        self.assertEqual(u'identity', element.name)
        self.assertEqual(u'conference', element.getAttribute(u'category'))
        self.assertEqual(u'text', element.getAttribute(u'type'))
        self.assertFalse(element.hasAttribute(u'name'))


    def test_fromElement(self):
        """
        Test creating L{disco.DiscoIdentity} from L{domish.Element}.
        """
        element = domish.Element((NS_DISCO_INFO, u'identity'))
        element['category'] = u'conference'
        element['type'] = u'text'
        element['name'] = u'The chatroom'
        identity = disco.DiscoIdentity.fromElement(element)
        self.assertEqual(u'conference', identity.category)
        self.assertEqual(u'text', identity.type)
        self.assertEqual(u'The chatroom', identity.name)


    def test_fromElementWithoutName(self):
        """
        Test creating L{disco.DiscoIdentity} from L{domish.Element}, no name.
        """
        element = domish.Element((NS_DISCO_INFO, u'identity'))
        element['category'] = u'conference'
        element['type'] = u'text'
        identity = disco.DiscoIdentity.fromElement(element)
        self.assertEqual(u'conference', identity.category)
        self.assertEqual(u'text', identity.type)
        self.assertEqual(None, identity.name)



class DiscoInfoTest(unittest.TestCase):
    """
    Tests for L{disco.DiscoInfo}.
    """

    def test_toElement(self):
        """
        Test C{toElement} creates a correctly namespaced element, no node.
        """
        info = disco.DiscoInfo()
        element = info.toElement()

        self.assertEqual(NS_DISCO_INFO, element.uri)
        self.assertEqual(u'query', element.name)
        self.assertFalse(element.hasAttribute(u'node'))


    def test_toElementNode(self):
        """
        Test C{toElement} with a node.
        """
        info = disco.DiscoInfo()
        info.nodeIdentifier = u'test'
        element = info.toElement()

        self.assertEqual(u'test', element.getAttribute(u'node'))


    def test_toElementChildren(self):
        """
        Test C{toElement} creates a DOM with proper childs.
        """
        info = disco.DiscoInfo()
        info.append(disco.DiscoFeature(u'jabber:iq:register'))
        info.append(disco.DiscoIdentity(u'conference', u'text'))
        info.append(data_form.Form(u'result'))
        element = info.toElement()

        featureElements = domish.generateElementsQNamed(element.children,
                                                        u'feature',
                                                        NS_DISCO_INFO)
        self.assertEqual(1, len(list(featureElements)))

        identityElements = domish.generateElementsQNamed(element.children,
                                                         u'identity',
                                                         NS_DISCO_INFO)
        self.assertEqual(1, len(list(identityElements)))

        extensionElements = domish.generateElementsQNamed(element.children,
                                                         u'x',
                                                         data_form.NS_X_DATA)
        self.assertEqual(1, len(list(extensionElements)))


    def test_fromElement(self):
        """
        Test properties when creating L{disco.DiscoInfo} from L{domish.Element}.
        """
        xml = """<query xmlns='http://jabber.org/protocol/disco#info'>
                   <identity category='conference'
                             type='text'
                             name='A Dark Cave'/>
                   <feature var='http://jabber.org/protocol/muc'/>
                   <feature var='jabber:iq:register'/>
                   <x xmlns='jabber:x:data' type='result'>
                     <field var='FORM_TYPE' type='hidden'>
                       <value>http://jabber.org/protocol/muc#roominfo</value>
                     </field>
                   </x>
                 </query>"""

        element = parseXml(xml)
        info = disco.DiscoInfo.fromElement(element)

        self.assertIn(u'http://jabber.org/protocol/muc', info.features)
        self.assertIn(u'jabber:iq:register', info.features)

        self.assertIn((u'conference', u'text'), info.identities)
        self.assertEqual(u'A Dark Cave',
                          info.identities[(u'conference', u'text')])

        self.assertIn(u'http://jabber.org/protocol/muc#roominfo',
                      info.extensions)


    def test_fromElementItems(self):
        """
        Test items when creating L{disco.DiscoInfo} from L{domish.Element}.
        """
        xml = """<query xmlns='http://jabber.org/protocol/disco#info'>
                   <identity category='conference'
                             type='text'
                             name='A Dark Cave'/>
                   <feature var='http://jabber.org/protocol/muc'/>
                   <feature var='jabber:iq:register'/>
                   <x xmlns='jabber:x:data' type='result'>
                     <field var='FORM_TYPE' type='hidden'>
                       <value>http://jabber.org/protocol/muc#roominfo</value>
                     </field>
                   </x>
                 </query>"""

        element = parseXml(xml)
        info = disco.DiscoInfo.fromElement(element)

        info = list(info)
        self.assertEqual(4, len(info))

        identity = info[0]
        self.assertEqual(u'conference', identity.category)

        self.assertEqual(u'http://jabber.org/protocol/muc', info[1])
        self.assertEqual(u'jabber:iq:register', info[2])

        extension = info[3]
        self.assertEqual(u'http://jabber.org/protocol/muc#roominfo',
                         extension.formNamespace)


    def test_fromElementNoNode(self):
        """
        Test creating L{disco.DiscoInfo} from L{domish.Element}, no node.
        """
        xml = """<query xmlns='http://jabber.org/protocol/disco#info'/>"""

        element = parseXml(xml)
        info = disco.DiscoInfo.fromElement(element)

        self.assertEqual(u'', info.nodeIdentifier)


    def test_fromElementNode(self):
        """
        Test creating L{disco.DiscoInfo} from L{domish.Element}, with node.
        """
        xml = """<query xmlns='http://jabber.org/protocol/disco#info'
                        node='test'>
                 </query>"""

        element = parseXml(xml)
        info = disco.DiscoInfo.fromElement(element)

        self.assertEqual(u'test', info.nodeIdentifier)



class DiscoItemTest(unittest.TestCase):
    """
    Tests for L{disco.DiscoItem}.
    """

    def test_init(self):
        """
        Test initialization with a category, type and name.
        """
        item = disco.DiscoItem(JID(u'example.org'), u'test', u'The node')
        self.assertEqual(JID(u'example.org'), item.entity)
        self.assertEqual(u'test', item.nodeIdentifier)
        self.assertEqual(u'The node', item.name)


    def test_toElement(self):
        """
        Test proper rendering to a DOM representation.

        The returned element should be properly named and have C{jid}, C{node},
        and C{name} attributes.
        """
        item = disco.DiscoItem(JID(u'example.org'), u'test', u'The node')
        element = item.toElement()
        self.assertEqual(NS_DISCO_ITEMS, element.uri)
        self.assertEqual(u'item', element.name)
        self.assertEqual(u'example.org', element.getAttribute(u'jid'))
        self.assertEqual(u'test', element.getAttribute(u'node'))
        self.assertEqual(u'The node', element.getAttribute(u'name'))


    def test_toElementWithoutName(self):
        """
        Test proper rendering to a DOM representation without a name.

        The returned element should be properly named and have C{jid}, C{node}
        attributes, no C{name} attribute.
        """
        item = disco.DiscoItem(JID(u'example.org'), u'test')
        element = item.toElement()
        self.assertEqual(NS_DISCO_ITEMS, element.uri)
        self.assertEqual(u'item', element.name)
        self.assertEqual(u'example.org', element.getAttribute(u'jid'))
        self.assertEqual(u'test', element.getAttribute(u'node'))
        self.assertFalse(element.hasAttribute(u'name'))


    def test_fromElement(self):
        """
        Test creating L{disco.DiscoItem} from L{domish.Element}.
        """
        element = domish.Element((NS_DISCO_ITEMS, u'item'))
        element[u'jid'] = u'example.org'
        element[u'node'] = u'test'
        element[u'name'] = u'The node'
        item = disco.DiscoItem.fromElement(element)
        self.assertEqual(JID(u'example.org'), item.entity)
        self.assertEqual(u'test', item.nodeIdentifier)
        self.assertEqual(u'The node', item.name)

    def test_fromElementNoNode(self):
        """
        Test creating L{disco.DiscoItem} from L{domish.Element}, no node.
        """
        element = domish.Element((NS_DISCO_ITEMS, u'item'))
        element[u'jid'] = u'example.org'
        element[u'name'] = u'The node'
        item = disco.DiscoItem.fromElement(element)
        self.assertEqual(JID(u'example.org'), item.entity)
        self.assertEqual(u'', item.nodeIdentifier)
        self.assertEqual(u'The node', item.name)


    def test_fromElementNoName(self):
        """
        Test creating L{disco.DiscoItem} from L{domish.Element}, no name.
        """
        element = domish.Element((NS_DISCO_ITEMS, u'item'))
        element[u'jid'] = u'example.org'
        element[u'node'] = u'test'
        item = disco.DiscoItem.fromElement(element)
        self.assertEqual(JID(u'example.org'), item.entity)
        self.assertEqual(u'test', item.nodeIdentifier)
        self.assertEqual(None, item.name)

    def test_fromElementBadJID(self):
        """
        Test creating L{disco.DiscoItem} from L{domish.Element}, bad JID.
        """
        element = domish.Element((NS_DISCO_ITEMS, u'item'))
        element[u'jid'] = u'ex@@@ample.org'
        item = disco.DiscoItem.fromElement(element)
        self.assertIdentical(None, item.entity)



class DiscoItemsTest(unittest.TestCase):
    """
    Tests for L{disco.DiscoItems}.
    """

    def test_toElement(self):
        """
        Test C{toElement} creates a correctly namespaced element, no node.
        """
        items = disco.DiscoItems()
        element = items.toElement()

        self.assertEqual(NS_DISCO_ITEMS, element.uri)
        self.assertEqual(u'query', element.name)
        self.assertFalse(element.hasAttribute(u'node'))


    def test_toElementNode(self):
        """
        Test C{toElement} with a node.
        """
        items = disco.DiscoItems()
        items.nodeIdentifier = u'test'
        element = items.toElement()

        self.assertEqual(u'test', element.getAttribute(u'node'))


    def test_toElementChildren(self):
        """
        Test C{toElement} creates a DOM with proper childs.
        """
        items = disco.DiscoItems()
        items.append(disco.DiscoItem(JID(u'example.org'), u'test', u'A node'))
        element = items.toElement()

        itemElements = domish.generateElementsQNamed(element.children,
                                                     u'item',
                                                     NS_DISCO_ITEMS)
        self.assertEqual(1, len(list(itemElements)))


    def test_fromElement(self):
        """
        Test creating L{disco.DiscoItems} from L{domish.Element}.
        """
        xml = """<query xmlns='http://jabber.org/protocol/disco#items'>
                   <item jid='example.org' node='test' name='A node'/>
                 </query>"""

        element = parseXml(xml)
        items = disco.DiscoItems.fromElement(element)

        items = list(items)
        self.assertEqual(1, len(items))
        item = items[0]

        self.assertEqual(JID(u'example.org'), item.entity)
        self.assertEqual(u'test', item.nodeIdentifier)
        self.assertEqual(u'A node', item.name)


    def test_fromElementNoNode(self):
        """
        Test creating L{disco.DiscoItems} from L{domish.Element}, no node.
        """
        xml = """<query xmlns='http://jabber.org/protocol/disco#items'/>"""

        element = parseXml(xml)
        items = disco.DiscoItems.fromElement(element)

        self.assertEqual(u'', items.nodeIdentifier)


    def test_fromElementNode(self):
        """
        Test creating L{disco.DiscoItems} from L{domish.Element}, with node.
        """
        xml = """<query xmlns='http://jabber.org/protocol/disco#items'
                        node='test'>
                 </query>"""

        element = parseXml(xml)
        items = disco.DiscoItems.fromElement(element)

        self.assertEqual(u'test', items.nodeIdentifier)



class DiscoClientProtocolTest(unittest.TestCase):
    """
    Tests for L{disco.DiscoClientProtocol}.
    """

    def setUp(self):
        """
        Set up stub and protocol for testing.
        """
        self.stub = XmlStreamStub()
        self.patch(XMPPHandler, 'request', self.request)
        self.protocol = disco.DiscoClientProtocol()


    def request(self, request):
        element = request.toElement()
        self.stub.xmlstream.send(element)
        return defer.Deferred()


    def test_requestItems(self):
        """
        Test request sent out by C{requestItems} and parsing of response.
        """
        def cb(items):
            items = list(items)
            self.assertEqual(2, len(items))
            self.assertEqual(JID(u'test.example.org'), items[0].entity)

        d = self.protocol.requestItems(JID(u'example.org'),u"foo")
        d.addCallback(cb)

        iq = self.stub.output[-1]
        self.assertEqual(u'example.org', iq.getAttribute(u'to'))
        self.assertEqual(u'get', iq.getAttribute(u'type'))
        self.assertEqual(u'foo', iq.query.getAttribute(u'node'))
        self.assertEqual(NS_DISCO_ITEMS, iq.query.uri)

        response = toResponse(iq, u'result')
        query = response.addElement((NS_DISCO_ITEMS, u'query'))

        element = query.addElement(u'item')
        element[u'jid'] = u'test.example.org'
        element[u'node'] = u'music'
        element[u'name'] = u'Music from the time of Shakespeare'

        element = query.addElement(u'item')
        element[u'jid'] = u"test2.example.org"

        d.callback(response)
        return d


    def test_requestItemsFrom(self):
        """
        A disco items request can be sent with an explicit sender address.
        """
        d = self.protocol.requestItems(JID(u'example.org'),
                                       sender=JID(u'test.example.org'))

        iq = self.stub.output[-1]
        self.assertEqual(u'test.example.org', iq.getAttribute(u'from'))

        response = toResponse(iq, u'result')
        response.addElement((NS_DISCO_ITEMS, u'query'))

        d.callback(response)
        return d


    def test_requestInfo(self):
        """
        Test request sent out by C{requestInfo} and parsing of response.
        """
        def cb(info):
            self.assertIn((u'conference', u'text'), info.identities)
            self.assertIn(u'http://jabber.org/protocol/disco#info',
                          info.features)
            self.assertIn(u'http://jabber.org/protocol/muc',
                          info.features)

        d = self.protocol.requestInfo(JID(u'example.org'),'foo')
        d.addCallback(cb)

        iq = self.stub.output[-1]
        self.assertEqual(u'example.org', iq.getAttribute(u'to'))
        self.assertEqual(u'get', iq.getAttribute(u'type'))
        self.assertEqual(u'foo', iq.query.getAttribute(u'node'))
        self.assertEqual(NS_DISCO_INFO, iq.query.uri)

        response = toResponse(iq, u'result')
        query = response.addElement((NS_DISCO_INFO, u'query'))

        element = query.addElement(u"identity")
        element[u'category'] = u'conference' # required
        element[u'type'] = u'text' # required
        element[u"name"] = u'Romeo and Juliet, Act II, Scene II' # optional

        element = query.addElement("feature")
        element[u'var'] = u'http://jabber.org/protocol/disco#info' # required

        element = query.addElement(u"feature")
        element[u'var'] = u'http://jabber.org/protocol/muc'

        d.callback(response)
        return d


    def test_requestInfoFrom(self):
        """
        A disco info request can be sent with an explicit sender address.
        """
        d = self.protocol.requestInfo(JID(u'example.org'),
                                      sender=JID(u'test.example.org'))

        iq = self.stub.output[-1]
        self.assertEqual(u'test.example.org', iq.getAttribute(u'from'))

        response = toResponse(iq, u'result')
        response.addElement((NS_DISCO_INFO, u'query'))

        d.callback(response)
        return d



class DiscoHandlerTest(unittest.TestCase, TestableRequestHandlerMixin):
    """
    Tests for L{disco.DiscoHandler}.
    """

    def setUp(self):
        self.service = disco.DiscoHandler()


    def test_connectionInitializedObserveInfo(self):
        """
        An observer for Disco Info requests is setup on stream initialization.
        """
        xml = """<iq from='test@example.com' to='example.com'
                     type='get'>
                   <query xmlns='%s'/>
                 </iq>""" % NS_DISCO_INFO

        def handleRequest(iq):
            called.append(iq)

        called = []
        self.service.xmlstream = utility.EventDispatcher()
        self.service.handleRequest = handleRequest
        self.service.connectionInitialized()
        self.service.xmlstream.dispatch(parseXml(xml))
        self.assertEqual(1, len(called))


    def test_connectionInitializedObserveItems(self):
        """
        An observer for Disco Items requests is setup on stream initialization.
        """
        xml = """<iq from='test@example.com' to='example.com'
                     type='get'>
                   <query xmlns='%s'/>
                 </iq>""" % NS_DISCO_ITEMS

        def handleRequest(iq):
            called.append(iq)

        called = []
        self.service.xmlstream = utility.EventDispatcher()
        self.service.handleRequest = handleRequest
        self.service.connectionInitialized()
        self.service.xmlstream.dispatch(parseXml(xml))
        self.assertEqual(1, len(called))


    def test_onDiscoInfo(self):
        """
        C{onDiscoInfo} should process an info request and return a response.

        The request should be parsed, C{info} called with the extracted
        parameters, and then the result should be formatted into a proper
        response element.
        """
        xml = """<iq from='test@example.com' to='example.com'
                     type='get'>
                   <query xmlns='%s'/>
                 </iq>""" % NS_DISCO_INFO

        def cb(element):
            self.assertEqual('query', element.name)
            self.assertEqual(NS_DISCO_INFO, element.uri)
            self.assertEqual(NS_DISCO_INFO, element.identity.uri)
            self.assertEqual('dummy', element.identity['category'])
            self.assertEqual('generic', element.identity['type'])
            self.assertEqual('Generic Dummy Entity', element.identity['name'])
            self.assertEqual(NS_DISCO_INFO, element.feature.uri)
            self.assertEqual('jabber:iq:version', element.feature['var'])

        def info(requestor, target, nodeIdentifier):
            self.assertEqual(JID('test@example.com'), requestor)
            self.assertEqual(JID('example.com'), target)
            self.assertEqual('', nodeIdentifier)

            return defer.succeed([
                disco.DiscoIdentity('dummy', 'generic', 'Generic Dummy Entity'),
                disco.DiscoFeature('jabber:iq:version')
            ])

        self.service.info = info
        d = self.handleRequest(xml)
        d.addCallback(cb)
        return d


    def test_onDiscoInfoWithNoFromAttribute(self):
        """
        Disco info request without a from attribute has requestor None.
        """
        xml = """<iq to='example.com'
                     type='get'>
                   <query xmlns='%s'/>
                 </iq>""" % NS_DISCO_INFO

        def info(requestor, target, nodeIdentifier):
            self.assertEqual(None, requestor)

            return defer.succeed([
                disco.DiscoIdentity('dummy', 'generic', 'Generic Dummy Entity'),
                disco.DiscoFeature('jabber:iq:version')
            ])

        self.service.info = info
        d = self.handleRequest(xml)
        return d


    def test_onDiscoInfoWithNoToAttribute(self):
        """
        Disco info request without a to attribute has target None.
        """
        xml = """<iq from='test@example.com'
                     type='get'>
                   <query xmlns='%s'/>
                 </iq>""" % NS_DISCO_INFO

        def info(requestor, target, nodeIdentifier):
            self.assertEqual(JID('test@example.com'), requestor)

            return defer.succeed([
                disco.DiscoIdentity('dummy', 'generic', 'Generic Dummy Entity'),
                disco.DiscoFeature('jabber:iq:version')
            ])

        self.service.info = info
        d = self.handleRequest(xml)
        return d


    def test_onDiscoInfoWithNode(self):
        """
        An info request for a node should return it in the response.
        """
        xml = """<iq from='test@example.com' to='example.com'
                     type='get'>
                   <query xmlns='%s' node='test'/>
                 </iq>""" % NS_DISCO_INFO

        def cb(element):
            self.assertTrue(element.hasAttribute('node'))
            self.assertEqual('test', element['node'])

        def info(requestor, target, nodeIdentifier):
            self.assertEqual('test', nodeIdentifier)

            return defer.succeed([
                disco.DiscoFeature('jabber:iq:version')
            ])

        self.service.info = info
        d = self.handleRequest(xml)
        d.addCallback(cb)
        return d


    def test_onDiscoInfoWithNodeNoResults(self):
        """
        An info request for a node with no results returns items-not-found.
        """
        xml = """<iq from='test@example.com' to='example.com'
                     type='get'>
                   <query xmlns='%s' node='test'/>
                 </iq>""" % NS_DISCO_INFO

        def cb(exc):
            self.assertEquals('item-not-found', exc.condition)

        def info(requestor, target, nodeIdentifier):
            self.assertEqual('test', nodeIdentifier)

            return defer.succeed([])

        self.service.info = info
        d = self.handleRequest(xml)
        self.assertFailure(d, StanzaError)
        d.addCallback(cb)
        return d


    def test_onDiscoItems(self):
        """
        C{onDiscoItems} should process an items request and return a response.

        The request should be parsed, C{items} called with the extracted
        parameters, and then the result should be formatted into a proper
        response element.
        """
        xml = """<iq from='test@example.com' to='example.com'
                     type='get'>
                   <query xmlns='%s'/>
                 </iq>""" % NS_DISCO_ITEMS

        def cb(element):
            self.assertEqual('query', element.name)
            self.assertEqual(NS_DISCO_ITEMS, element.uri)
            self.assertEqual(NS_DISCO_ITEMS, element.item.uri)
            self.assertEqual('example.com', element.item['jid'])
            self.assertEqual('test', element.item['node'])
            self.assertEqual('Test node', element.item['name'])

        def items(requestor, target, nodeIdentifier):
            self.assertEqual(JID('test@example.com'), requestor)
            self.assertEqual(JID('example.com'), target)
            self.assertEqual('', nodeIdentifier)

            return defer.succeed([
                disco.DiscoItem(JID('example.com'), 'test', 'Test node'),
            ])

        self.service.items = items
        d = self.handleRequest(xml)
        d.addCallback(cb)
        return d


    def test_onDiscoItemsWithNode(self):
        """
        An items request for a node should return it in the response.
        """
        xml = """<iq from='test@example.com' to='example.com'
                     type='get'>
                   <query xmlns='%s' node='test'/>
                 </iq>""" % NS_DISCO_ITEMS

        def cb(element):
            self.assertTrue(element.hasAttribute('node'))
            self.assertEqual('test', element['node'])

        def items(requestor, target, nodeIdentifier):
            self.assertEqual('test', nodeIdentifier)

            return defer.succeed([
                disco.DiscoFeature('jabber:iq:version')
            ])

        self.service.items = items
        d = self.handleRequest(xml)
        d.addCallback(cb)
        return d


    def test_info(self):
        """
        C{info} should gather disco info from sibling handlers.
        """
        discoItems = [disco.DiscoIdentity('dummy', 'generic',
                                          'Generic Dummy Entity'),
                      disco.DiscoFeature('jabber:iq:version')
        ]

        class DiscoResponder(XMPPHandler):
            implements(disco.IDisco)

            def getDiscoInfo(self, requestor, target, nodeIdentifier):
                if not nodeIdentifier:
                    return defer.succeed(discoItems)
                else:
                    return defer.succeed([])

        def cb(result):
            self.assertEquals(discoItems, result)

        self.service.parent = [self.service, DiscoResponder()]
        d = self.service.info(JID('test@example.com'), JID('example.com'), '')
        d.addCallback(cb)
        return d


    def test_infoNotDeferred(self):
        """
        C{info} should gather disco info from sibling handlers.
        """
        discoItems = [disco.DiscoIdentity('dummy', 'generic',
                                          'Generic Dummy Entity'),
                      disco.DiscoFeature('jabber:iq:version')
        ]

        class DiscoResponder(XMPPHandler):
            implements(disco.IDisco)

            def getDiscoInfo(self, requestor, target, nodeIdentifier):
                if not nodeIdentifier:
                    return discoItems
                else:
                    return []

        def cb(result):
            self.assertEquals(discoItems, result)

        self.service.parent = [self.service, DiscoResponder()]
        d = self.service.info(JID('test@example.com'), JID('example.com'), '')
        d.addCallback(cb)
        return d


    def test_items(self):
        """
        C{info} should gather disco items from sibling handlers.
        """
        discoItems = [disco.DiscoItem(JID('example.com'), 'test', 'Test node')]

        class DiscoResponder(XMPPHandler):
            implements(disco.IDisco)

            def getDiscoItems(self, requestor, target, nodeIdentifier):
                if not nodeIdentifier:
                    return defer.succeed(discoItems)
                else:
                    return defer.succeed([])

        def cb(result):
            self.assertEquals(discoItems, result)

        self.service.parent = [self.service, DiscoResponder()]
        d = self.service.items(JID('test@example.com'), JID('example.com'), '')
        d.addCallback(cb)
        return d


    def test_itemsNotDeferred(self):
        """
        C{info} should also collect results not returned via a deferred.
        """
        discoItems = [disco.DiscoItem(JID('example.com'), 'test', 'Test node')]

        class DiscoResponder(XMPPHandler):
            implements(disco.IDisco)

            def getDiscoItems(self, requestor, target, nodeIdentifier):
                if not nodeIdentifier:
                    return discoItems
                else:
                    return []

        def cb(result):
            self.assertEquals(discoItems, result)

        self.service.parent = [self.service, DiscoResponder()]
        d = self.service.items(JID('test@example.com'), JID('example.com'), '')
        d.addCallback(cb)
        return d



# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Unit test helpers.
"""

from twisted.internet import defer
from twisted.words.xish import xpath
from twisted.words.xish.utility import EventDispatcher

from wokkel.generic import parseXml
from wokkel.subprotocols import StreamManager

class XmlStreamStub(object):
    """
    Stub for testing objects that communicate through XML Streams.

    Instances of this stub hold an object in L{xmlstream} that acts like an
    L{XmlStream<twisted.words.xish.xmlstream.XmlStream} after connection stream
    initialization. Stanzas can be sent through it by calling its C{send}
    method with an object implementing
    L{IElement<twisted.words.xish.domish.IElement>} as its first argument.
    These appear in sequence in the L{output} instance variable of the stub.

    For the reverse direction, stanzas passed to L{send} of the stub, will be
    dispatched in the stubbed XmlStream as if it was received over the wire, so
    that registered observers will be called.

    Example:

        >>> stub = XmlStreamStub()
        >>> stub.xmlstream.send(domish.Element((None, 'presence')))
        >>> stub.output[-1].toXml()
        u'<presence/>'
        >>> def cb(stanza):
        ...     print "Got: %r" stanza.toXml()
        >>> stub.xmlstream.addObserver('/presence')
        >>> stub.send(domish.Element((None, 'presence')))
        Got: u'<presence/>'

    @ivar xmlstream: Stubbed XML Stream.
    @type xmlstream: L{EventDispatcher}
    @ivar output: List of stanzas sent to the XML Stream.
    @type output: L{list}
    """

    def __init__(self):
        self.output = []
        self.xmlstream = EventDispatcher()
        self.xmlstream.send = self.output.append

    def send(self, obj):
        """
        Pass an element to the XML Stream as if received.

        @param obj: Element to be dispatched to C{self.xmlstream}.
        @type obj: object implementing
                   L{IElement<twisted.words.xish.domish.IElement>}.
        """
        self.xmlstream.dispatch(obj)


class TestableRequestHandlerMixin(object):
    """
    Mixin for testing XMPPHandlers that process iq requests.

    Handlers that use L{wokkel.subprotocols.IQHandlerMixin} define a
    C{iqHandlers} attribute that lists the handlers to be called for iq
    requests. This mixin provides L{handleRequest} to mimic the handler
    processing for easier testing.
    """

    def handleRequest(self, xml):
        """
        Find a handler and call it directly.

        @param xml: XML stanza that may yield a handler being called.
        @type xml: C{str}.
        @return: Deferred that fires with the result of a handler for this
                 stanza. If no handler was found, the deferred has its errback
                 called with a C{NotImplementedError} exception.
        """
        handler = None
        iq = parseXml(xml)
        for queryString, method in self.service.iqHandlers.iteritems():
            if xpath.internQuery(queryString).matches(iq):
                handler = getattr(self.service, method)

        if handler:
            d = defer.maybeDeferred(handler, iq)
        else:
            d = defer.fail(NotImplementedError())

        return d


class TestableStreamManager(StreamManager):
    """
    Stream manager for testing subprotocol handlers.
    """

    def __init__(self, reactor=None):
        class DummyFactory(object):
            def addBootstrap(self, event, fn):
                pass

        factory = DummyFactory()
        StreamManager.__init__(self, factory, reactor)
        self.stub = XmlStreamStub()
        self._connected(self.stub.xmlstream)
        self._authd(self.stub.xmlstream)



"""
An XMPP subprotocol handler that acts as an XMPP Ping pinger.
"""

from wokkel.ping import PingClientProtocol

class Pinger(PingClientProtocol):
    """
    I send a ping as soon as I have a connection.
    """

    def __init__(self, entity, sender=None):
        self.entity = entity
        self.sender = sender

    def connectionInitialized(self):
        def cb(response):
            print "*** Pong ***"

        print "*** Ping ***"
        d = self.ping(self.entity, sender=self.sender)
        d.addCallback(cb)



#!/usr/bin/env python

# Copyright (c) Ralph Meijer.
# See LICENSE for details.

from setuptools import setup

# Make sure 'twisted' doesn't appear in top_level.txt

try:
    from setuptools.command import egg_info
    egg_info.write_toplevel_names
except (ImportError, AttributeError):
    pass
else:
    def _top_level_package(name):
        return name.split('.', 1)[0]

    def _hacked_write_toplevel_names(cmd, basename, filename):
        pkgs = dict.fromkeys(
            [_top_level_package(k)
                for k in cmd.distribution.iter_distribution_names()
                if _top_level_package(k) != "twisted"
            ]
        )
        cmd.write_file("top-level names", filename, '\n'.join(pkgs) + '\n')

    egg_info.write_toplevel_names = _hacked_write_toplevel_names

setup(name='wokkel',
      version='0.7.1',
      description='Twisted Jabber support library',
      author='Ralph Meijer',
      author_email='ralphm@ik.nu',
      maintainer_email='ralphm@ik.nu',
      url='http://wokkel.ik.nu/',
      license='MIT',
      platforms='any',
      packages=[
          'wokkel',
          'wokkel.test',
          'twisted.plugins',
      ],
      package_data={'twisted.plugins': ['twisted/plugins/server.py']},
      zip_safe=False,
      install_requires=[
          'Twisted >= 10.0.0',
          'python-dateutil',
      ],
)



# Copyright (c) Twisted Matrix Laboratories.
# Copyright (c) Ralph Meijer.
# See LICENSE for details.

"""
Tests for L{wokkel.compat}.
"""

from zope.interface import implements
from twisted.internet import task
from twisted.internet.interfaces import IReactorTime
from twisted.trial import unittest
from twisted.words.protocols.jabber import xmlstream

from wokkel.compat import IQ
from wokkel.compat import NamedConstant, Names, ValueConstant, Values

class DeprecationTest(unittest.TestCase):
    """
    Deprecation tests for L{wokkel.compat}.
    """

    def lookForDeprecationWarning(self, testmethod, attributeName, newName):
        """
        Importing C{testmethod} emits a deprecation warning.
        """
        warningsShown = self.flushWarnings([testmethod])
        self.assertEqual(len(warningsShown), 1)
        self.assertIdentical(warningsShown[0]['category'], DeprecationWarning)
        self.assertEqual(
            warningsShown[0]['message'],
            "wokkel.compat." + attributeName + " "
            "was deprecated in Wokkel 0.7.0: Use " + newName + " instead.")


    def test_bootstrapMixinTest(self):
        """
        L{compat.BootstrapMixin} is deprecated.
        """
        from wokkel.compat import BootstrapMixin
        BootstrapMixin
        self.lookForDeprecationWarning(
                self.test_bootstrapMixinTest,
                "BootstrapMixin",
                "twisted.words.xish.xmlstream.BootstrapMixin")


    def test_xmlStreamServerFactory(self):
        """
        L{compat.XmlStreamServerFactory} is deprecated.
        """
        from wokkel.compat import XmlStreamServerFactory
        XmlStreamServerFactory
        self.lookForDeprecationWarning(
                self.test_xmlStreamServerFactory,
                "XmlStreamServerFactory",
                "twisted.words.protocols.jabber.xmlstream."
                    "XmlStreamServerFactory")



class FakeReactor(object):

    implements(IReactorTime)
    def __init__(self):
        self.clock = task.Clock()
        self.callLater = self.clock.callLater
        self.getDelayedCalls = self.clock.getDelayedCalls



class IQTest(unittest.TestCase):
    """
    Tests for L{IQ}.
    """

    def setUp(self):
        self.reactor = FakeReactor()
        self.clock = self.reactor.clock


    def testRequestTimingOutEventDispatcher(self):
        """
        Test that an iq request with a defined timeout times out.
        """
        from twisted.words.xish import utility
        output = []
        xs = utility.EventDispatcher()
        xs.send = output.append

        self.iq = IQ(xs, reactor=self.reactor)
        self.iq.timeout = 60
        d = self.iq.send()
        self.assertFailure(d, xmlstream.TimeoutError)

        self.clock.pump([1, 60])
        self.assertFalse(self.reactor.getDelayedCalls())
        self.assertFalse(xs.iqDeferreds)
        return d



class NamedConstantTests(unittest.TestCase):
    """
    Tests for the L{twisted.python.constants.NamedConstant} class which is used
    to represent individual values.
    """
    def setUp(self):
        """
        Create a dummy container into which constants can be placed.
        """
        class foo(Names):
            pass
        self.container = foo


    def test_name(self):
        """
        The C{name} attribute of a L{NamedConstant} refers to the value passed
        for the C{name} parameter to C{_realize}.
        """
        name = NamedConstant()
        name._realize(self.container, "bar", None)
        self.assertEqual("bar", name.name)


    def test_representation(self):
        """
        The string representation of an instance of L{NamedConstant} includes
        the container the instances belongs to as well as the instance's name.
        """
        name = NamedConstant()
        name._realize(self.container, "bar", None)
        self.assertEqual("<foo=bar>", repr(name))


    def test_equality(self):
        """
        A L{NamedConstant} instance compares equal to itself.
        """
        name = NamedConstant()
        name._realize(self.container, "bar", None)
        self.assertTrue(name == name)
        self.assertFalse(name != name)


    def test_nonequality(self):
        """
        Two different L{NamedConstant} instances do not compare equal to each
        other.
        """
        first = NamedConstant()
        first._realize(self.container, "bar", None)
        second = NamedConstant()
        second._realize(self.container, "bar", None)
        self.assertFalse(first == second)
        self.assertTrue(first != second)


    def test_hash(self):
        """
        Because two different L{NamedConstant} instances do not compare as equal
        to each other, they also have different hashes to avoid collisions when
        added to a C{dict} or C{set}.
        """
        first = NamedConstant()
        first._realize(self.container, "bar", None)
        second = NamedConstant()
        second._realize(self.container, "bar", None)
        self.assertNotEqual(hash(first), hash(second))



class _ConstantsTestsMixin(object):
    """
    Mixin defining test helpers common to multiple types of constants
    collections.
    """
    def _notInstantiableTest(self, name, cls):
        """
        Assert that an attempt to instantiate the constants class raises
        C{TypeError}.

        @param name: A C{str} giving the name of the constants collection.
        @param cls: The constants class to test.
        """
        exc = self.assertRaises(TypeError, cls)
        self.assertEqual(name + " may not be instantiated.", str(exc))



class NamesTests(unittest.TestCase, _ConstantsTestsMixin):
    """
    Tests for L{twisted.python.constants.Names}, a base class for containers of
    related constaints.
    """
    def setUp(self):
        """
        Create a fresh new L{Names} subclass for each unit test to use.  Since
        L{Names} is stateful, re-using the same subclass across test methods
        makes exercising all of the implementation code paths difficult.
        """
        class METHOD(Names):
            """
            A container for some named constants to use in unit tests for
            L{Names}.
            """
            GET = NamedConstant()
            PUT = NamedConstant()
            POST = NamedConstant()
            DELETE = NamedConstant()

        self.METHOD = METHOD


    def test_notInstantiable(self):
        """
        A subclass of L{Names} raises C{TypeError} if an attempt is made to
        instantiate it.
        """
        self._notInstantiableTest("METHOD", self.METHOD)


    def test_symbolicAttributes(self):
        """
        Each name associated with a L{NamedConstant} instance in the definition
        of a L{Names} subclass is available as an attribute on the resulting
        class.
        """
        self.assertTrue(hasattr(self.METHOD, "GET"))
        self.assertTrue(hasattr(self.METHOD, "PUT"))
        self.assertTrue(hasattr(self.METHOD, "POST"))
        self.assertTrue(hasattr(self.METHOD, "DELETE"))


    def test_withoutOtherAttributes(self):
        """
        As usual, names not defined in the class scope of a L{Names}
        subclass are not available as attributes on the resulting class.
        """
        self.assertFalse(hasattr(self.METHOD, "foo"))


    def test_representation(self):
        """
        The string representation of a constant on a L{Names} subclass includes
        the name of the L{Names} subclass and the name of the constant itself.
        """
        self.assertEqual("<METHOD=GET>", repr(self.METHOD.GET))


    def test_lookupByName(self):
        """
        Constants can be looked up by name using L{Names.lookupByName}.
        """
        method = self.METHOD.lookupByName("GET")
        self.assertIdentical(self.METHOD.GET, method)


    def test_notLookupMissingByName(self):
        """
        Names not defined with a L{NamedConstant} instance cannot be looked up
        using L{Names.lookupByName}.
        """
        self.assertRaises(ValueError, self.METHOD.lookupByName, "lookupByName")
        self.assertRaises(ValueError, self.METHOD.lookupByName, "__init__")
        self.assertRaises(ValueError, self.METHOD.lookupByName, "foo")


    def test_name(self):
        """
        The C{name} attribute of one of the named constants gives that
        constant's name.
        """
        self.assertEqual("GET", self.METHOD.GET.name)


    def test_attributeIdentity(self):
        """
        Repeated access of an attribute associated with a L{NamedConstant} value
        in a L{Names} subclass results in the same object.
        """
        self.assertIdentical(self.METHOD.GET, self.METHOD.GET)


    def test_iterconstants(self):
        """
        L{Names.iterconstants} returns an iterator over all of the constants
        defined in the class, in the order they were defined.
        """
        constants = list(self.METHOD.iterconstants())
        self.assertEqual(
            [self.METHOD.GET, self.METHOD.PUT,
             self.METHOD.POST, self.METHOD.DELETE],
            constants)


    def test_attributeIterconstantsIdentity(self):
        """
        The constants returned from L{Names.iterconstants} are identical to the
        constants accessible using attributes.
        """
        constants = list(self.METHOD.iterconstants())
        self.assertIdentical(self.METHOD.GET, constants[0])
        self.assertIdentical(self.METHOD.PUT, constants[1])
        self.assertIdentical(self.METHOD.POST, constants[2])
        self.assertIdentical(self.METHOD.DELETE, constants[3])


    def test_iterconstantsIdentity(self):
        """
        The constants returned from L{Names.iterconstants} are identical on each
        call to that method.
        """
        constants = list(self.METHOD.iterconstants())
        again = list(self.METHOD.iterconstants())
        self.assertIdentical(again[0], constants[0])
        self.assertIdentical(again[1], constants[1])
        self.assertIdentical(again[2], constants[2])
        self.assertIdentical(again[3], constants[3])


    def test_initializedOnce(self):
        """
        L{Names._enumerants} is initialized once and its value re-used on
        subsequent access.
        """
        first = self.METHOD._enumerants
        self.METHOD.GET # Side-effects!
        second = self.METHOD._enumerants
        self.assertIdentical(first, second)



class ValuesTests(unittest.TestCase, _ConstantsTestsMixin):
    """
    Tests for L{twisted.python.constants.Names}, a base class for containers of
    related constaints with arbitrary values.
    """
    def setUp(self):
        """
        Create a fresh new L{Values} subclass for each unit test to use.  Since
        L{Values} is stateful, re-using the same subclass across test methods
        makes exercising all of the implementation code paths difficult.
        """
        class STATUS(Values):
            OK = ValueConstant("200")
            NOT_FOUND = ValueConstant("404")

        self.STATUS = STATUS


    def test_notInstantiable(self):
        """
        A subclass of L{Values} raises C{TypeError} if an attempt is made to
        instantiate it.
        """
        self._notInstantiableTest("STATUS", self.STATUS)


    def test_symbolicAttributes(self):
        """
        Each name associated with a L{ValueConstant} instance in the definition
        of a L{Values} subclass is available as an attribute on the resulting
        class.
        """
        self.assertTrue(hasattr(self.STATUS, "OK"))
        self.assertTrue(hasattr(self.STATUS, "NOT_FOUND"))


    def test_withoutOtherAttributes(self):
        """
        As usual, names not defined in the class scope of a L{Values}
        subclass are not available as attributes on the resulting class.
        """
        self.assertFalse(hasattr(self.STATUS, "foo"))


    def test_representation(self):
        """
        The string representation of a constant on a L{Values} subclass includes
        the name of the L{Values} subclass and the name of the constant itself.
        """
        self.assertEqual("<STATUS=OK>", repr(self.STATUS.OK))


    def test_lookupByName(self):
        """
        Constants can be looked up by name using L{Values.lookupByName}.
        """
        method = self.STATUS.lookupByName("OK")
        self.assertIdentical(self.STATUS.OK, method)


    def test_notLookupMissingByName(self):
        """
        Names not defined with a L{ValueConstant} instance cannot be looked up
        using L{Values.lookupByName}.
        """
        self.assertRaises(ValueError, self.STATUS.lookupByName, "lookupByName")
        self.assertRaises(ValueError, self.STATUS.lookupByName, "__init__")
        self.assertRaises(ValueError, self.STATUS.lookupByName, "foo")


    def test_lookupByValue(self):
        """
        Constants can be looked up by their associated value, defined by the
        argument passed to L{ValueConstant}, using L{Values.lookupByValue}.
        """
        status = self.STATUS.lookupByValue("200")
        self.assertIdentical(self.STATUS.OK, status)


    def test_lookupDuplicateByValue(self):
        """
        If more than one constant is associated with a particular value,
        L{Values.lookupByValue} returns whichever of them is defined first.
        """
        class TRANSPORT_MESSAGE(Values):
            """
            Message types supported by an SSH transport.
            """
            KEX_DH_GEX_REQUEST_OLD = ValueConstant(30)
            KEXDH_INIT = ValueConstant(30)

        self.assertIdentical(
            TRANSPORT_MESSAGE.lookupByValue(30),
            TRANSPORT_MESSAGE.KEX_DH_GEX_REQUEST_OLD)


    def test_notLookupMissingByValue(self):
        """
        L{Values.lookupByValue} raises L{ValueError} when called with a value
        with which no constant is associated.
        """
        self.assertRaises(ValueError, self.STATUS.lookupByValue, "OK")
        self.assertRaises(ValueError, self.STATUS.lookupByValue, 200)
        self.assertRaises(ValueError, self.STATUS.lookupByValue, "200.1")


    def test_name(self):
        """
        The C{name} attribute of one of the constants gives that constant's
        name.
        """
        self.assertEqual("OK", self.STATUS.OK.name)


    def test_attributeIdentity(self):
        """
        Repeated access of an attribute associated with a L{ValueConstant} value
        in a L{Values} subclass results in the same object.
        """
        self.assertIdentical(self.STATUS.OK, self.STATUS.OK)


    def test_iterconstants(self):
        """
        L{Values.iterconstants} returns an iterator over all of the constants
        defined in the class, in the order they were defined.
        """
        constants = list(self.STATUS.iterconstants())
        self.assertEqual(
            [self.STATUS.OK, self.STATUS.NOT_FOUND],
            constants)


    def test_attributeIterconstantsIdentity(self):
        """
        The constants returned from L{Values.iterconstants} are identical to the
        constants accessible using attributes.
        """
        constants = list(self.STATUS.iterconstants())
        self.assertIdentical(self.STATUS.OK, constants[0])
        self.assertIdentical(self.STATUS.NOT_FOUND, constants[1])


    def test_iterconstantsIdentity(self):
        """
        The constants returned from L{Values.iterconstants} are identical on
        each call to that method.
        """
        constants = list(self.STATUS.iterconstants())
        again = list(self.STATUS.iterconstants())
        self.assertIdentical(again[0], constants[0])
        self.assertIdentical(again[1], constants[1])


    def test_initializedOnce(self):
        """
        L{Values._enumerants} is initialized once and its value re-used on
        subsequent access.
        """
        first = self.STATUS._enumerants
        self.STATUS.OK # Side-effects!
        second = self.STATUS._enumerants
        self.assertIdentical(first, second)
